// This file is licensed under the Elastic License 2.0. Copyright 2021-present, StarRocks Limited.

#include "storage/vectorized/chunk_helper.h"

#include "column/binary_column.h"
#include "column/chunk.h"
#include "column/column.h"
#include "column/field.h"
#include "column/nullable_column.h"
#include "column/schema.h"
#include "gtest/gtest.h"
#include "runtime/descriptor_helper.h"
#include "storage/schema.h"
#include "util/logging.h"

namespace starrocks {
namespace vectorized {

class ChunkHelperTest : public testing::Test {
public:
    void add_tablet_column(TabletSchemaPB& tablet_schema_pb, int32_t id, bool is_key, std::string type, int32_t length,
                           bool is_nullable);
    starrocks::Schema* gen_schema(bool is_nullable);
    vectorized::SchemaPtr gen_v_schema(bool is_nullable);
    void check_chunk(Chunk* chunk, size_t column_size, size_t row_size);
    void check_chunk_nullable(Chunk* chunk, size_t column_size, size_t row_size);
    void check_column(Column* column, FieldType type, size_t row_size);

private:
    FieldType _type[9] = {OLAP_FIELD_TYPE_TINYINT, OLAP_FIELD_TYPE_SMALLINT, OLAP_FIELD_TYPE_INT,
                          OLAP_FIELD_TYPE_BIGINT,  OLAP_FIELD_TYPE_LARGEINT, OLAP_FIELD_TYPE_FLOAT,
                          OLAP_FIELD_TYPE_DOUBLE,  OLAP_FIELD_TYPE_VARCHAR,  OLAP_FIELD_TYPE_CHAR};

    PrimitiveType _primitive_type[9] = {
            PrimitiveType::TYPE_TINYINT, PrimitiveType::TYPE_SMALLINT, PrimitiveType::TYPE_INT,
            PrimitiveType::TYPE_BIGINT,  PrimitiveType::TYPE_LARGEINT, PrimitiveType::TYPE_FLOAT,
            PrimitiveType::TYPE_DOUBLE,  PrimitiveType::TYPE_VARCHAR,  PrimitiveType::TYPE_CHAR};

    TSlotDescriptor _create_slot_desc(PrimitiveType type, const std::string& col_name, int col_pos);
    TupleDescriptor* _create_tuple_desc();

    ObjectPool _pool;
};

TSlotDescriptor ChunkHelperTest::_create_slot_desc(PrimitiveType type, const std::string& col_name, int col_pos) {
    TSlotDescriptorBuilder builder;

    if (type == PrimitiveType::TYPE_VARCHAR || type == PrimitiveType::TYPE_CHAR) {
        return builder.string_type(1024).column_name(col_name).column_pos(col_pos).nullable(false).build();
    } else {
        return builder.type(type).column_name(col_name).column_pos(col_pos).nullable(false).build();
    }
}

TupleDescriptor* ChunkHelperTest::_create_tuple_desc() {
    TDescriptorTableBuilder table_builder;
    TTupleDescriptorBuilder tuple_builder;

    for (size_t i = 0; i < 9; i++) {
        tuple_builder.add_slot(_create_slot_desc(_primitive_type[i], "c" + std::to_string(i), 0));
    }

    tuple_builder.build(&table_builder);

    std::vector<TTupleId> row_tuples = std::vector<TTupleId>{0};
    std::vector<bool> nullable_tuples = std::vector<bool>{true};
    DescriptorTbl* tbl = nullptr;
    DescriptorTbl::create(&_pool, table_builder.desc_tbl(), &tbl, config::vector_chunk_size);

    auto* row_desc = _pool.add(new RowDescriptor(*tbl, row_tuples, nullable_tuples));
    auto* tuple_desc = row_desc->tuple_descriptors()[0];

    return tuple_desc;
}

void ChunkHelperTest::add_tablet_column(TabletSchemaPB& tablet_schema_pb, int32_t id, bool is_key, std::string type,
                                        int32_t length, bool is_nullable) {
    ColumnPB* column = tablet_schema_pb.add_column();
    column->set_unique_id(id);
    column->set_name("c" + std::to_string(id));
    column->set_type(type);
    column->set_is_key(is_key);
    column->set_length(length);
    column->set_is_nullable(is_nullable);
    column->set_aggregation("NONE");
}

starrocks::Schema* ChunkHelperTest::gen_schema(bool is_nullable) {
    TabletSchemaPB tablet_schema_pb;
    add_tablet_column(tablet_schema_pb, 0, true, "TINYINT", 1, is_nullable);
    add_tablet_column(tablet_schema_pb, 1, false, "SMALLINT", 2, is_nullable);
    add_tablet_column(tablet_schema_pb, 2, false, "INT", 4, is_nullable);
    add_tablet_column(tablet_schema_pb, 3, false, "BIGINT", 8, is_nullable);
    add_tablet_column(tablet_schema_pb, 4, false, "LARGEINT", 16, is_nullable);
    add_tablet_column(tablet_schema_pb, 5, false, "FLOAT", 4, is_nullable);
    add_tablet_column(tablet_schema_pb, 6, false, "DOUBLE", 8, is_nullable);
    add_tablet_column(tablet_schema_pb, 7, false, "VARCHAR", 16, is_nullable);
    add_tablet_column(tablet_schema_pb, 8, false, "CHAR", 16, is_nullable);

    TabletSchema* tablet_schema = new TabletSchema();
    tablet_schema->init_from_pb(tablet_schema_pb);
    return new starrocks::Schema(*tablet_schema);
}

vectorized::SchemaPtr ChunkHelperTest::gen_v_schema(bool is_nullable) {
    vectorized::Fields fields;
    fields.emplace_back(std::make_shared<Field>(0, "c0", get_type_info(OLAP_FIELD_TYPE_TINYINT), is_nullable));
    fields.emplace_back(std::make_shared<Field>(1, "c1", get_type_info(OLAP_FIELD_TYPE_SMALLINT), is_nullable));
    fields.emplace_back(std::make_shared<Field>(2, "c2", get_type_info(OLAP_FIELD_TYPE_INT), is_nullable));
    fields.emplace_back(std::make_shared<Field>(3, "c3", get_type_info(OLAP_FIELD_TYPE_BIGINT), is_nullable));
    fields.emplace_back(std::make_shared<Field>(4, "c4", get_type_info(OLAP_FIELD_TYPE_LARGEINT), is_nullable));
    fields.emplace_back(std::make_shared<Field>(5, "c5", get_type_info(OLAP_FIELD_TYPE_FLOAT), is_nullable));
    fields.emplace_back(std::make_shared<Field>(6, "c6", get_type_info(OLAP_FIELD_TYPE_DOUBLE), is_nullable));
    fields.emplace_back(std::make_shared<Field>(7, "c7", get_type_info(OLAP_FIELD_TYPE_VARCHAR), is_nullable));
    fields.emplace_back(std::make_shared<Field>(8, "c8", get_type_info(OLAP_FIELD_TYPE_CHAR), is_nullable));
    return std::make_shared<Schema>(fields);
}

void ChunkHelperTest::check_chunk(Chunk* chunk, size_t column_size, size_t row_size) {
    CHECK_EQ(chunk->columns().size(), column_size);
    for (size_t i = 0; i < column_size; i++) {
        check_column(chunk->get_column_by_index(i).get(), _type[i], row_size);
    }
}

void ChunkHelperTest::check_chunk_nullable(Chunk* chunk, size_t column_size, size_t row_size) {
    CHECK_EQ(chunk->columns().size(), column_size);
    for (size_t i = 0; i < column_size; i++) {
        Column* d_column =
                (reinterpret_cast<NullableColumn*>(chunk->get_column_by_index(i).get()))->data_column().get();
        check_column(d_column, _type[i], row_size);
    }
}

void ChunkHelperTest::check_column(Column* column, FieldType type, size_t row_size) {
    ASSERT_EQ(column->size(), row_size);

    switch (type) {
    case OLAP_FIELD_TYPE_TINYINT: {
        const int8_t* data = reinterpret_cast<const int8_t*>(static_cast<Int8Column*>(column)->raw_data());
        for (int i = 0; i < row_size; i++) {
            ASSERT_EQ(*(data + i), static_cast<int8_t>(i * 2));
        }
        break;
    }
    case OLAP_FIELD_TYPE_SMALLINT: {
        const int16_t* data = reinterpret_cast<const int16_t*>(static_cast<Int16Column*>(column)->raw_data());
        for (int i = 0; i < row_size; i++) {
            ASSERT_EQ(*(data + i), static_cast<int16_t>(i * 2 * 10));
        }
        break;
    }
    case OLAP_FIELD_TYPE_INT: {
        const int32_t* data = reinterpret_cast<const int32_t*>(static_cast<Int32Column*>(column)->raw_data());
        for (int i = 0; i < row_size; i++) {
            ASSERT_EQ(*(data + i), static_cast<int32_t>(i * 2 * 100));
        }
        break;
    }
    case OLAP_FIELD_TYPE_BIGINT: {
        const int64_t* data = reinterpret_cast<const int64_t*>(static_cast<Int64Column*>(column)->raw_data());
        for (int i = 0; i < row_size; i++) {
            ASSERT_EQ(*(data + i), static_cast<int64_t>(i * 2 * 1000));
        }
        break;
    }
    case OLAP_FIELD_TYPE_LARGEINT: {
        const int128_t* data = reinterpret_cast<const int128_t*>(static_cast<Int128Column*>(column)->raw_data());
        for (int i = 0; i < row_size; i++) {
            ASSERT_EQ(*(data + i), static_cast<int128_t>(i * 2 * 10000));
        }
        break;
    }
    case OLAP_FIELD_TYPE_FLOAT: {
        const float* data = reinterpret_cast<const float*>(static_cast<FloatColumn*>(column)->raw_data());
        for (int i = 0; i < row_size; i++) {
            ASSERT_EQ(*(data + i), static_cast<float>(i * 2 * 100000));
        }
        break;
    }
    case OLAP_FIELD_TYPE_DOUBLE: {
        const double* data = reinterpret_cast<const double*>(static_cast<DoubleColumn*>(column)->raw_data());
        for (int i = 0; i < row_size; i++) {
            ASSERT_EQ(*(data + i), static_cast<double>(i * 2 * 1000000));
        }
        break;
    }
    case OLAP_FIELD_TYPE_VARCHAR: {
        const BinaryColumn* data = reinterpret_cast<const BinaryColumn*>(column);
        for (int i = 0; i < row_size; i++) {
            Slice l = data->get_slice(i);
            Slice r(std::to_string(i * 2 * 10000000));
            ASSERT_EQ(l, r);
        }
        break;
    }
    case OLAP_FIELD_TYPE_CHAR: {
        const BinaryColumn* data = reinterpret_cast<const BinaryColumn*>(column);
        for (int i = 0; i < row_size; i++) {
            Slice l = data->get_slice(i);
            Slice r(std::to_string(i * 2 * 100000000));
            ASSERT_EQ(l, r);
        }
        break;
    }
    default:
        break;
    }
}

TEST_F(ChunkHelperTest, NewChunkWithTuple) {
    auto* tuple_desc = _create_tuple_desc();

    auto chunk = ChunkHelper::new_chunk(*tuple_desc, 1024);

    // check
    ASSERT_EQ(chunk->num_columns(), 9);
    ASSERT_EQ(chunk->get_column_by_slot_id(0)->get_name(), "integral-1");
    ASSERT_EQ(chunk->get_column_by_slot_id(1)->get_name(), "integral-2");
    ASSERT_EQ(chunk->get_column_by_slot_id(2)->get_name(), "integral-4");
    ASSERT_EQ(chunk->get_column_by_slot_id(3)->get_name(), "integral-8");
    ASSERT_EQ(chunk->get_column_by_slot_id(4)->get_name(), "int128");
    ASSERT_EQ(chunk->get_column_by_slot_id(5)->get_name(), "float-4");
    ASSERT_EQ(chunk->get_column_by_slot_id(6)->get_name(), "float-8");
    ASSERT_EQ(chunk->get_column_by_slot_id(7)->get_name(), "binary");
    ASSERT_EQ(chunk->get_column_by_slot_id(8)->get_name(), "binary");
}

} // namespace vectorized
} // namespace starrocks
